from .fixtures import *  # required for all
import os
import yaml
import subprocess
from ..consts import gitcmd as git
from ..repo import Repo
from . import temppath
from .tools import _make_remote_repo
from .tools import clone_and_commit
from .tools import gimera_apply


def test_switch_submodule_to_integrated_migrate_changes(temppath):
    """
    A gimera sub has changes and is submodule.
    Changes shall not be lost when switching to integrated modus.
    """

    def _make_changes():
        # make it dirty
        dirty_file = workspace_main / "a" / "b" / "sub1" / "file1.txt"
        original_content = dirty_file.read_text()
        dirty_file.write_text("i changed the file")
        # add a file
        added_file = workspace_main / "a" / "b" / "sub1" / "newfile.txt"
        added_file.write_text("new file")
        # delete a file
        deleted_file = workspace_main / "a" / "b" / "sub1" / "dont_look_at_me"
        deleted_file.unlink()

    def _assert_changes():
        dirty_file = workspace_main / "a" / "b" / "sub1" / "file1.txt"
        added_file = workspace_main / "a" / "b" / "sub1" / "newfile.txt"
        deleted_file = workspace_main / "a" / "b" / "sub1" / "dont_look_at_me"
        assert dirty_file.read_text() == "i changed the file"
        assert added_file.exists()
        assert not deleted_file.exists()

    workspace = temppath / "test_switch_submodule_to_integrated_migrate_changes"
    workspace.mkdir()
    workspace_main = workspace / "main_working"

    repo_main = _make_remote_repo(temppath / "mainrepo")
    repo_sub = _make_remote_repo(temppath / "sub1")

    repos_sub = {
        "repos": [
            {
                "url": f"file://{repo_sub}",
                "branch": "branch1",
                "path": "a/b/sub1",
                "patches": [],
                "type": "submodule",
            },
        ]
    }
    with clone_and_commit(repo_sub, "branch1") as repopath:
        (repopath / "repo_sub.txt").write_text("This is a new function")
        (repopath / "dont_look_at_me").write_text("i am ugly")
        Repo(repopath).simple_commit_all()

    subprocess.check_output(
        git + ["clone", "file://" + str(repo_main), workspace_main],
        cwd=workspace.parent,
    )
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_sub))
    (workspace_main / "main.txt").write_text("main repo")
    repo = Repo(workspace_main)
    repo.simple_commit_all()
    repo.X(*(git + ["push"]))

    os.chdir(workspace_main)
    gimera_apply([], None)

    _make_changes()

    os.chdir(workspace_main)
    gimera_apply([], update=True, migrate_changes=True)

    # check if still dirty
    _assert_changes()



from .fixtures import * # required for all
import uuid
import yaml
from contextlib import contextmanager
from ..repo import Repo
import os
import subprocess
import click
import inspect
import os
from pathlib import Path
from .tools import gimera_apply
from . import temppath
from .tools import _make_remote_repo
from .tools import clone_and_commit
from .tools import gimera_commit

from ..consts import gitcmd as git

def test_checkout_not_update_if_last_commit_matches_branch_make_branch_be_checked_out(
    temppath,
):
    workspace = temppath / "workspace_checkout_match_branch"
    workspace.mkdir(exist_ok=True)
    os.chdir(workspace.parent)

    repo_main = _make_remote_repo(workspace / "mainrepo")
    repo_submodule = _make_remote_repo(workspace / "repo_submodule")

    with clone_and_commit(repo_submodule, "main") as repopath:
        (repopath / "submodule.txt").write_text("This is a new function")
        repo = Repo(repopath)
        repo.simple_commit_all()
        sha = repo.out("git", "log", "-n1", "--format=%H")
        (repopath / "submodule.txt").write_text("This is a new function2")
        repo.simple_commit_all()

    repos = {
        "repos": [
            {
                "url": f"file://{repo_submodule}",
                "branch": "main",
                "path": "sub1",
                "sha": sha,
                "type": "submodule",
            },
        ]
    }
    workspace_main = workspace / "main_working"
    main_repo = Repo(workspace_main)
    subprocess.check_call(git + ["clone", f"file://{repo_main}", workspace_main])
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos))

    os.chdir(workspace_main)
    gimera_apply([], update=None, recursive=True, strict=True)

    assert (workspace_main / "sub1" / "submodule.txt").exists()

    def _get_branch():
        branch = [
            x
            for x in subprocess.check_output(
                ["git", "branch"], encoding="utf8", cwd=workspace_main / "sub1"
            )
            .strip()
            .splitlines()
            if x.startswith("* ")
        ][0]
        return branch

    assert sha[:7] in _get_branch()

    os.chdir(workspace_main)
    gimera_apply([], update=True)
    assert " main" in _get_branch()



import re
import os
from inquirer import errors
import shutil
import sys
import subprocess
import click
from .repo import Repo
from datetime import datetime
from contextlib import contextmanager
from .consts import gitcmd as git
from .tools import confirm
from .tools import temppath
from .tools import path1inpath2
from .tools import verbose
from .consts import inquirer_theme
from .tools import (
    _raise_error,
    get_nearest_repo,
    temppath,
    rsync,
    wait_git_lock,
    files_relative_to,
    filter_files_to_folders,
)
import inquirer
from pathlib import Path
from .consts import REPO_TYPE_INT, REPO_TYPE_SUB
from .config import Config


def make_patches(working_dir, main_repo, repo_yml, common_vars):
    if repo_yml.type != REPO_TYPE_INT:
        raise NotImplementedError(repo_yml.type)
    verbose(f"Making patches for {repo_yml.path}")

    with _if_ignored_move_to_separate_dir(
        working_dir, main_repo, repo_yml, common_vars
    ) as (main_repo, is_temp_path):
        with _prepare(main_repo, repo_yml) as (
            subrepo,
            subrepo_path,
            changed_files,
            untracked_files,
        ):
            if not changed_files:
                return
            if not _start_question(repo_yml, changed_files):
                return

            with _temporarily_add_untracked_files(main_repo, untracked_files):
                subdir_path = Path(main_repo.path) / repo_yml.path
                with main_repo.temporary_unignore(subdir_path):
                    patch_content = _technically_make_patch(main_repo, subdir_path)

                    if not repo_yml.patches:
                        if os.getenv("GIMERA_FORCE") == "1":
                            return
                        if os.getenv("GIMERA_NON_INTERACTIVE") != "1" and not is_temp_path:
                            repo_yml = _ask_user_to_create_path_directory(repo_yml)
                        if not repo_yml.patches:
                            _raise_error(
                                "Please define at least one directory, "
                                f"where patches are stored for {repo_yml.path}"
                            )

                with _prepare_patchdir(repo_yml) as (patch_dir, patch_filename):
                    _write_patch_content(
                        main_repo,
                        repo_yml,
                        subrepo,
                        subrepo_path,
                        patch_dir,
                        patch_filename,
                        patch_content,
                    )


def _ask_user_to_create_path_directory(repo_yml):
    def validation(answers, current):
        if current and current.startswith("/"):
            raise errors.ValidationError(
                "", reason="Please provide relative paths only."
            )
        return True

    questions = [
        inquirer.Text(
            "path",
            validate=validation,
            message=(
                f"Please define a patch directory for {repo_yml.path} "
                " - relative path please starting from root."
            ),
        )
    ]
    answers = inquirer.prompt(questions, theme=inquirer_theme)
    path = answers["path"]
    from .config import PatchDir

    repo_yml.config._store(repo_yml, {"patches": [PatchDir(repo_yml, path, None)]})
    repo_yml.config.load_config()
    repo_yml = [x for x in repo_yml.config.repos if x.path == repo_yml.path][0]
    return repo_yml


@contextmanager
def _temporarily_move_gimera(repo_yml, to_path):
    remember_config_path = repo_yml.config.config_file
    repo_yml.config.config_file = to_path / "gimera.yml"
    repo_yml.config.config_file.parent.mkdir(exist_ok=True, parents=True)
    shutil.copy(remember_config_path, repo_yml.config.config_file)

    yield

    repo_yml.config.config_file = remember_config_path


@contextmanager
def _if_ignored_move_to_separate_dir(working_dir, main_repo, repo_yml, common_vars):
    """
    If directory is ignored then move to temporary path.
    Apply changes from local dir to get the diffs.

    also transfer the parent .gitignore file, so that e.g. pyc files are 
    ignored
    """
    from .cachedir import _get_cache_dir
    from .integrated import _update_integrated_module

    if (
        main_repo.check_ignore(repo_yml.path)
        and (main_repo.path / repo_yml.path).exists()
    ):
        with temppath() as path:
            subprocess.check_call(
                (git + ["init", "--initial-branch=main", "."]), cwd=path
            )
            Repo(path).X(*(git + ["commit", "-m", 'init', '--allow-empty']), output=True)
            main_repo2 = Repo(path)
            for patchdir in repo_yml.patches:
                dest_path = main_repo2.path / patchdir._path
                dest_path.mkdir(exist_ok=True, parents=True)
                patch_path = main_repo.path / patchdir._path
                if not patch_path.exists():
                    if os.getenv("GIMERA_NON_INTERACTIVE") != "1":
                        confirm(f"Path {patch_path} does not exist. Create?")
                        patch_path.mkdir(parents=True, exist_ok=True)
                rsync(main_repo.path / patchdir._path, dest_path)
            main_repo2.simple_commit_all()
            with _temporarily_move_gimera(repo_yml, main_repo2.path):
                _update_integrated_module(
                    main_repo2.path,
                    main_repo2,
                    repo_yml,
                    update=False,
                    common_vars=common_vars,
                )
                main_repo2.simple_commit_all()

                # now transfer the latest changes:
                rsync(
                    main_repo.path / repo_yml.path,
                    path / repo_yml.path,
                    exclude=[".git"],
                )
                gitignore_file = main_repo.path / ".gitignore"
                if gitignore_file.exists():
                    shutil.copy(
                        gitignore_file,
                        path / ".gitignore",
                    )
                yield main_repo2, True

                for patchdir in repo_yml.patches:
                    rsync(main_repo2.path / patchdir._path, main_repo.path / patchdir._path)
    else:
        yield main_repo, False


def _update_edited_patchfile(repo_yml):
    click.secho(
        "Editing a patch is in progress - continuing for " f"{repo_yml.edit_patchfile}",
        fg="yellow",
    )
    ttype = repo_yml._get_type_of_patchfolder(Path(repo_yml.edit_patchfile).parent)
    if (ttype) == "from_outside":
        edit_patchfile = repo_yml.config.config_file.parent / repo_yml.edit_patchfile
    elif ttype == "internal":
        edit_patchfile = (
            repo_yml.config.config_file.parent / repo_yml.path / repo_yml.edit_patchfile
        )
    else:
        raise NotImplementedError(ttype)
    patch_dir = [
        x
        for x in repo_yml.patches
        if x.path_absolute == edit_patchfile.parent
    ][0]
    patch_filename = Path(repo_yml.edit_patchfile).name
    return patch_dir, patch_filename


def _get_new_patchfilename(repo_yml):
    patchdirs = repo_yml.patches
    if len(patchdirs) == 1:
        patch_dir = patchdirs[0]
    else:
        if os.getenv("GIMERA_NON_INTERACTIVE") == "1":
            _raise_error(
                (
                    "A patch dir is required but non interactive mode is set. "
                    "You can provide the --no-patch option perhaps."
                )
            )

        questions = [
            inquirer.List(
                "path",
                message="Please choose a directory where to put the patch file.",
                # choices=["Type directory"] + patchdirs,
                choices=patchdirs,
            )
        ]
        answers = inquirer.prompt(questions, theme=inquirer_theme)
        patch_dir = answers["path"]
    return patch_dir


@contextmanager
def _prepare_patchdir(repo_yml):
    remove_edit_patchfile = False
    if repo_yml.edit_patchfile:
        patch_dir, patch_filename = _update_edited_patchfile(repo_yml)
        remove_edit_patchfile = True
    else:
        patch_filename = datetime.now().strftime("%Y%m%d_%H%M%S")
        patch_dir = _get_new_patchfilename(repo_yml)

    patch_dir._path.mkdir(exist_ok=True, parents=True)

    if os.getenv("GIMERA_NON_INTERACTIVE") != "1" and not repo_yml.edit_patchfile:
        questions = [
            inquirer.Text(
                "filename",
                message="Please give the patch-file a name",
            )
        ]
        answers = inquirer.prompt(questions, theme=inquirer_theme)
        if not answers:
            sys.exit(-1)
        patch_filename = answers["filename"]
    if not patch_filename:
        _raise_error("No filename provided")

    if not patch_filename.endswith(".patch"):
        patch_filename += ".patch"

    yield patch_dir, patch_filename

    if remove_edit_patchfile:
        repo_yml.config._store(
            repo_yml,
            {
                "edit_patchfile": "",
            },
        )


def _start_question(repo_yml, changed_files):
    files_in_lines = "\n".join(map(str, sorted(changed_files)))
    if os.getenv("GIMERA_NON_INTERACTIVE") == "1":
        correct = True
    else:
        choice_yes = "Yes - make a patch"
        if repo_yml.edit_patchfile:
            choice_yes = f"Merge all changes into patchfile {repo_yml.edit_patchfile}"

        for file in files_in_lines.splitlines():
            click.secho(f"  * {file}")

        questions = [
            inquirer.List(
                "correct",
                message=f"Continue making patches for the lines above?",
                default="no",
                choices=[
                    choice_yes,
                    "Abort",
                    "Ignore",
                ],
            )
        ]
        answers = inquirer.prompt(questions, theme=inquirer_theme)
        correct = answers["correct"][0]
        if correct == "Y":
            correct = True
        elif correct == "A":
            correct = False
    if not correct:
        sys.exit(-1)
    if correct == "I":
        return False
    return True


def _write_patch_content(
    main_repo, repo_yml, subrepo, subrepo_path, patch_dir, patch_filename, patch_content
):
    if path1inpath2(patch_dir._path, subrepo_path):
        # case: patch must be put within patches folder of the integrated module
        # so it must be uploaded via a temp path; and the latest version must be
        # pulled

        hex = _clone_directory_and_add_patch_file(
            main_repo=main_repo,
            repo_yml=repo_yml,
            branch=repo_yml.branch,
            repo_url=repo_yml.url,
            patch_path=patch_dir._path.relative_to(subrepo_path) / patch_filename,
            content=patch_content,
        )
        # write latest hex to gimera
        repo_yml.config._store(
            repo_yml,
            {
                "sha": hex,
            },
        )
        repo_yml.sha = hex

    else:
        # case: patch file is in main repo and can be committed there
        (patch_dir.path_absolute / patch_filename).write_text(patch_content)

        # commit the patches - do NOT - could lie in submodule - is hard to do
        with main_repo.temporary_unignore(repo_yml.path):
            # make sure by that, that e.g. pyc files are ignored
            subprocess.check_call((git + ["add", repo_yml.path]), cwd=main_repo.path)
        subprocess.check_call((git + ["add", patch_dir._path]), cwd=main_repo.path)
        subprocess.check_call(
            (git + ["commit", "-m", f"added patch {patch_filename}", "--no-verify"]),
            cwd=main_repo.path,
        )


@contextmanager
def _temporarily_add_untracked_files(repo, untracked_files):
    for untracked_file in untracked_files:
        # add with empty blob to index, appears like that then:
        #
        # Changes not staged for commit:
        # (use "git add <file>..." to update what will be committed)
        # (use "git restore <file>..." to discard changes in working directory)
        #         modified:   roles2/sub1/file2.txt
        #         new file:   roles2/sub1/file3.txt
        repo.X(*(git + ["add", "-f", "-N", untracked_file]))

    yield

    for to_reset in untracked_files:
        repo.X(*(git + ["reset", to_reset]))


@contextmanager
def _prepare(main_repo, repo_yml):
    subrepo_path = main_repo.path / repo_yml.path
    if not subrepo_path.exists():
        yield None, None, [], []
    else:
        try:
            subrepo = main_repo.get_submodule(repo_yml.path)
        except ValueError:
            subrepo = None

        if subrepo and subrepo.path.exists():
            changed_files = subrepo.all_dirty_files
            untracked_files = subrepo.untracked_files
        elif subrepo_path.exists():
            repo = Repo(get_nearest_repo(main_repo.path, subrepo_path))
            changed_files = repo.all_dirty_files_absolute
            untracked_files = repo.untracked_files_absolute
            changed_files = list(
                files_relative_to(
                    filter_files_to_folders(changed_files, [subrepo_path]), repo.path
                )
            )
            untracked_files = list(
                files_relative_to(
                    filter_files_to_folders(untracked_files, [subrepo_path]), repo.path
                )
            )

        else:
            changed_files, untracked_files = [], []
        yield subrepo, subrepo_path, changed_files, untracked_files


def _clone_directory_and_add_patch_file(
    main_repo, repo_yml, branch, repo_url, patch_path, content
):
    from .fetch import _fetch_branch, _get_cache_dir

    with temppath() as path:
        path = path / "repo"
        subprocess.check_call((git + ["clone", "--branch", branch, repo_url, path]))
        repo = Repo(path)
        repo.X(*(git + ["checkout", branch]))
        patch_path = path / patch_path
        assert patch_path.relative_to(path)
        patch_path.parent.mkdir(exist_ok=True, parents=True)
        patch_path.write_text(content)
        repo.X(*(git + ["add", patch_path.relative_to(path)]))
        repo.X(*(git + ["commit", "-m", f"added patchfile: {patch_path}"]))
        repo.X(*(git + ["push"]))
        del repo
        # also make sure that local cache is updated, because
        # latest repo version is applied to project
        with _get_cache_dir(main_repo, repo_yml) as cache_dir:
            with wait_git_lock(cache_dir):
                repo = Repo(cache_dir)
                _fetch_branch(repo, repo_yml, filter_remote="origin")
                with repo.worktree(branch) as repo:
                    repo.pull(repo_yml=repo_yml)
                    return repo.hex


def _technically_make_patch(repo, path):
    repo.X(*(git + ["add", path]))
    repo.X(*(git + ["commit", "-m", "for patch", "--no-verify"]))

    patch_content = Repo(path).out(
        *(git + ["format-patch", "HEAD~1", "--stdout", "--relative"])
    )
    repo.X(*(git + ["reset", "HEAD~1"]))
    return patch_content


def _apply_patches(repo_yml):
    verbose(f"Applying patches for {repo_yml.path}")
    relevant_patch_files = set()
    for patchdir in repo_yml.patches:
        # with patchdir.path as dir:
        if not patchdir.path_absolute.exists():
            patchdir.path_absolute.mkdir(parents=True)
        for file in sorted(patchdir.path_absolute.rglob("*.patch")):
            if repo_yml.ignore_patchfile(file):
                continue
            element = (patchdir, file)
            if not [x for x in relevant_patch_files if x[1] == file]:
                relevant_patch_files.add(element)
        del patchdir

    for patchdir, file in sorted(relevant_patch_files, key=lambda x: x[1].name):
        click.secho((f"Applying patch {file}"), fg="blue")
        # Git apply fails silently if applied within local repos
        _apply_patchfile(file, patchdir.apply_from_here_dir, error_ok=False)


def _apply_patchfile(file, working_dir, error_ok=False, just_check=False):
    verbose(f"Applying patchfile {file} in working dir {working_dir}")
    cwd = Path(working_dir)
    # must be check_output due to input keyword
    # Explaining -R option:
    #   at testing a patchfile is created ; although not comitting
    #   git detects, that the file was removed and same patch tries to be applied
    #   Very intelligent but we force defined state over such smart behaviours.
    """
    /tmp/gimeratest/workspace/integrated/sub1/patches/15.0/superpatches/my.patch
    =============================================================================================
    patching file file1.txt
    Reversed (or previously applied) patch detected!  Assume -R? [n]
    Apply anyway? [n]
    Skipping patch.
    1 out of 1 hunk ignored -- saving rejects to file file1.txt.rej
    """
    file = Path(file)
    try:
        cmd = [
            "patch",
            "-p1",
            "--no-backup-if-mismatch",
            "--force",
            "-s",
            "-i",
            str(file),
        ]
        if just_check:
            cmd += ["--dry-run"]
        subprocess.check_output(cmd, cwd=cwd, encoding="utf-8")
        click.secho(
            (f"Applied patch {file}"),
            fg="blue",
        )
    except subprocess.CalledProcessError as ex:
        click.secho(
            ("\n\nFailed to apply the following patch file:\n\n"),
            fg="yellow",
        )
        click.secho(
            (
                f"{file}\n"
                "============================================================================================="
            ),
            fg="red",
            bold=True,
        )
        click.secho((f"{ex.stdout or ''}\n" f"{ex.stderr or ''}\n"), fg="yellow")

        click.secho(file.read_text(), fg="cyan")
        if os.getenv("GIMERA_NON_INTERACTIVE") == "1" or not inquirer.confirm(
            f"Patchfile failed ''{file}'' - continue with next file?",
            default=True,
        ):
            if not error_ok:
                _raise_error(f"Error applying patch: {file}")
        return False
    except Exception as ex:  # pylint: disable=broad-except
        _raise_error(str(ex))
        return False
    else:
        return True


def _apply_patchfile(file, working_dir, error_ok=False):
    cwd = Path(working_dir)
    # must be check_output due to input keyword
    # Explaining -R option:
    #   at testing a patchfile is created ; although not comitting
    #   git detects, that the file was removed and same patch tries to be applied
    #   Very intelligent but we force defined state over such smart behaviours.
    """
    /tmp/gimeratest/workspace/integrated/sub1/patches/15.0/superpatches/my.patch
    =============================================================================================
    patching file file1.txt
    Reversed (or previously applied) patch detected!  Assume -R? [n]
    Apply anyway? [n]
    Skipping patch.
    1 out of 1 hunk ignored -- saving rejects to file file1.txt.rej
    """
    file = Path(file)
    try:
        subprocess.check_output(
            ["patch", "-p1", "--no-backup-if-mismatch", "--force", "-i", str(file)],
            cwd=cwd,
            encoding="utf-8",
        )
        click.secho(
            (f"Applied patch {file}"),
            fg="blue",
        )
    except subprocess.CalledProcessError as ex:
        click.secho(
            ("\n\nFailed to apply the following patch file:\n\n"),
            fg="yellow",
        )
        click.secho(
            (
                f"{file}\n"
                "============================================================================================="
            ),
            fg="red",
            bold=True,
        )
        click.secho((f"{ex.stdout or ''}\n" f"{ex.stderr or ''}\n"), fg="yellow")

        click.secho(file.read_text(), fg="cyan")
        if os.getenv("GIMERA_NON_INTERACTIVE") == "1" or not inquirer.confirm(
            f"Patchfile failed ''{file}'' - continue with next file?",
            default=True,
        ):
            if not error_ok:
                _raise_error(f"Error applying patch: {file}")
    except Exception as ex:  # pylint: disable=broad-except
        _raise_error(str(ex))


def _get_repo_to_patchfiles(patchfiles):
    for patchfile in patchfiles:
        patchfile = Path(patchfile)
        if patchfile.exists() and str(patchfile).startswith("/"):
            patchfile = str(patchfile.relative_to(Path(os.getcwd())))
        patchfile = _get_available_patchfiles(None, None, str(patchfile))
        if not patchfile:
            _raise_error(f"Not found: {patchfile}")
        if len(patchfile) > 1:
            _raise_error(f"Too many patchfiles found: {patchfile}")

        cwd = Path(os.getcwd())
        patchfile = cwd / patchfile[0]
        config = Config(force_type=False)

        def _get_repo_of_patchfile():
            for repo in config.repos:
                if not repo.enabled:
                    continue
                patches = repo.patches
                if not patches:
                    continue
                for patchdir in patches:
                    path = patchdir.path_absolute
                    for file in path.glob("*.patch"):
                        if file == patchfile:
                            return repo

        repo = _get_repo_of_patchfile()
        if not repo:
            _raise_error(f"Repo not found for {patchfile}")

        if repo.type != REPO_TYPE_INT:
            _raise_error(f"Repo {repo.path} is not integrated")
        yield (repo, patchfile)


def _edit_patch(patchfiles):
    from .apply import _internal_apply

    patchfiles = list(sorted(set(patchfiles)))
    for patchfile in list(_get_repo_to_patchfiles(patchfiles)):
        repo, patchfile = patchfile
        if repo.edit_patchfile:
            _raise_error(f"Already WIP for patchfile: {repo.edit_patchfile}")
        try:
            repo.edit_patchfile = str(patchfile.relative_to(repo.fullpath))
        except ValueError:
            repo.edit_patchfile = patchfile.relative_to(repo.config.config_file.parent)
        repo.config._store(
            repo,
            {
                "edit_patchfile": str(repo.edit_patchfile),
            },
        )
        break
    _internal_apply(str(repo.path), update=False, force_type=None)


def _get_available_patchfiles(ctx, param, incomplete):
    config = Config(force_type=False, recursive=True)
    cwd = Path(os.getcwd())
    patchfiles = []
    filtered_patchfiles = []
    for repo in config.repos:
        if not repo.enabled:
            continue
        for patchdir in repo.patches:
            if not patchdir.path_absolute.exists():
                continue
            for file in patchdir.path_absolute.glob("*.patch"):
                patchfiles.append(file.relative_to(cwd))
    if incomplete:
        for file in patchfiles:
            if incomplete in str(file):
                filtered_patchfiles.append(file)
    else:
        filtered_patchfiles = patchfiles
    filtered_patchfiles = list(sorted(filtered_patchfiles))
    return sorted(list(set(map(str, filtered_patchfiles))))


def remove_file_from_patch(files_to_exclude, patchfilecontent):
    if not patchfilecontent:
        return
    lines = patchfilecontent.split(b"\n")

    new_lines = []
    skip_lines = False

    for line in lines:
        line = line.decode("utf8")
        # Check if the line indicates the start of a new diff section
        if line.startswith("diff --git"):
            # Extract the file path from the diff line
            match = re.search(r"a/(.+) b/(.+)", line)
            if match:
                file_path = match.group(1)
                # Check if the file should be excluded
                if any(file_path.startswith(x) for x in files_to_exclude):
                    skip_lines = True
                else:
                    skip_lines = False

        # Only add lines if we're not skipping them
        if not skip_lines:
            new_lines.append(line)

    return b"\n".join(map(lambda x: x.encode("utf8"), new_lines))



import uuid
from datetime import datetime
import pytest
import os
import sys
import shutil
import tempfile
from pathlib import Path


@pytest.fixture(autouse=True)
def set_env_vars():
    os.environ["GIMERA_EXCEPTION_THAN_SYSEXIT"] = "1"
    os.environ["GIMERA_FORCE"] = "0"
    os.environ["GIMERA_NON_INTERACTIVE"] = "0"
    # otherwise test2 submodules fails; repos are fetched in threads;
    # just happens at tests; perhaps because of pytest - not in real world 
    # (test in console)
    os.environ["GIMERA_NON_THREADED"] = "1"


@pytest.fixture(autouse=True)
def python():
    return sys.executable


@pytest.fixture(autouse=True)
def temppath():
    dt = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
    path = Path(f"/tmp/gimeratest/{dt}{str(uuid.uuid4())[:5]}")
    path = Path(f"/tmp/gimeratest")
    if path.exists():
        shutil.rmtree(path)
    path.mkdir(exist_ok=True, parents=True)
    try:
        yield path
    finally:
        if path.exists():
            shutil.rmtree(path)


@pytest.fixture(autouse=True)
def cleangimera_cache():
    cache_dir = Path(os.path.expanduser("~")) / ".cache/gimera"
    backup_dir = cache_dir.parent / f"{cache_dir.name}_backup"
    if cache_dir.exists():
        if backup_dir.exists():
            shutil.rmtree(backup_dir)
        shutil.move(cache_dir, backup_dir)
    yield
    if cache_dir.exists():
        shutil.rmtree(cache_dir)
    if backup_dir.exists():
        shutil.move(backup_dir, cache_dir)



from .fixtures import * # required for all
import os
import subprocess
from ..consts import gitcmd as git
from ..repo import Repo
from .tools import _make_remote_repo
from .tools import clone_and_commit
from . import temppath


def test_git_status(temppath):
    """
    make dirty submodule then repo.full_clean
    """
    workspace = temppath / "workspace_git_status"
    os.chdir(workspace.parent)

    repo_main = _make_remote_repo(temppath / "mainrepo")

    with clone_and_commit(repo_main, "main") as repopath:
        (repopath / "file1.txt").write_text("This is a new function")
        Repo(repopath).simple_commit_all()

    workspace_main = workspace / "main_working"
    subprocess.check_call(git + ["clone", f"file://{repo_main}", workspace_main])
    (workspace_main / "file8.txt").write_text("Newfile")
    repo = Repo(workspace_main)
    assert not repo.staged_files
    assert repo.untracked_files

def test_reformat_url(temppath):
    from ..tools import get_url_type, reformat_url

    url = "git@github.com:marcwimmer/gimera.git"
    assert get_url_type(url) == "git"

    assert reformat_url(url, "http") == "https://github.com/marcwimmer/gimera.git"




import subprocess
from .tools import get_url_type, reformat_url
import traceback
import os
import threading
import click
import uuid
from .submodule import _has_repo_latest_commit
from .consts import gitcmd as git
from .repo import Repo
from .cachedir import _get_cache_dir
from .tools import verbose
from .tools import wait_git_lock
from .tools import _raise_error
from .tools import try_rm_tree
from .tools import assert_exception_no_exit

threadLimiter = threading.BoundedSemaphore(4)


def _fetch_repos_in_parallel(
    main_repo, repos, update=None, minimal_fetch=None, no_fetch=None
):
    results = {"errors": {}, "urls": set()}
    if os.getenv("GIMERA_NON_THREADED", "0") == "1":
        threaded = False
    else:
        threaded = len(repos) > 1

    def _pull_repo(index, main_repo, repo_yml):
        threadLimiter.acquire()
        try:
            if repo_yml.url in results["urls"]:
                return
            verbose(f"Fetching {repo_yml.url}")
            results["urls"].add(repo_yml.url)
            with _get_cache_dir(
                main_repo, repo_yml, no_action_if_not_exist=True
            ) as cache_dir:
                if cache_dir is not None:
                    repo = Repo(cache_dir)
                    do_fetch = True
                    if minimal_fetch:
                        with wait_git_lock(cache_dir):
                            if repo_yml.sha:
                                if repo.contains(repo_yml.sha):
                                    do_fetch = False
                            else:
                                if repo.contains_branch(repo_yml.branch):
                                    do_fetch = False

                    if do_fetch:
                        with wait_git_lock(cache_dir):
                            _fetch_branch(
                                repo, repo_yml, filter_remote="origin", no_fetch=False
                            )

        except Exception as ex:
            if os.getenv("GIMERA_IGNORE_FETCH_ERRORS") == "1":
                click.secho(
                    "Following error is ignored because GIMERA_IGNORE_FETCH_ERRORS is set:"
                )
                click.secho(str(ex), fg="red")
            else:
                trace = traceback.format_exc()
                results["errors"][main_repo.path] = f"{ex}\n\n{trace}"
                if not threaded:
                    raise
        finally:
            threadLimiter.release()

    if not threaded:
        for index, repo in enumerate(repos):
            _pull_repo(index, main_repo, repo)
    else:

        threads = []
        for index, repo in enumerate(repos):
            t = threading.Thread(target=_pull_repo, args=(index, main_repo, repo))
            t.daemon = True
            threads.append(t)
        [x.start() for x in threads]
        [x.join() for x in threads]

        if results["errors"]:
            raise Exception(results["errors"])


def _fetch_branch(repo, repo_yml, no_fetch=False, filter_remote=None, **options):
    url = repo_yml.url

    fetch_exception = None
    if no_fetch:
        return
    with assert_exception_no_exit():
        for remote in repo.remotes:
            try:
                url = remote.url
                _set_url_and_fetch(
                    repo, repo_yml, remote.name, url, filter_remote=filter_remote
                )
            except Exception as ex:
                fetch_exception = ex
                for combination in [
                    ("git", "http"),
                    ("http", "git"),
                ]:
                    if get_url_type(url) == combination[0]:
                        url_http = reformat_url(url, combination[1])
                        try:
                            _set_url_and_fetch(
                                repo,
                                repo_yml,
                                remote.name,
                                url_http,
                                filter_remote=filter_remote,
                            )
                            break
                        except Exception:
                            raise fetch_exception
                else:
                    raise fetch_exception


def _set_url_and_fetch(
    repo, repo_yml, remote_name, url, filter_remote=None, trycount=0
):
    repo.set_remote_url(remote_name, url)
    branch = repo_yml.branch
    todo_branches = [branch]
    success = False

    with wait_git_lock(repo.path):
        try:
            repo.out(*(git + ["fetch", remote_name] + todo_branches))
            for branch in todo_branches:
                remote_sha = (
                    repo.X(*(git + ["ls-remote", "origin", branch]), output=True)
                    .strip()
                    .split("\t")[0]
                )
                repo.X(*(git + ["update-ref", f"refs/heads/{branch}", remote_sha]))
            success = True
        except subprocess.CalledProcessError as ex:
            click.secho(ex.stderr, fg="red")

    if success:
        if not _has_repo_latest_commit(repo, repo_yml.branch):
            success = False

    if not success:
        if trycount == 0:
            click.secho(
                (
                    f"This the absolutely LAST RESORT: deleting now {repo.path} as "
                    "fetching the origin and updating the branches did not update the "
                    "local branches."
                )
            )

            try_rm_tree(repo.path)
            with _get_cache_dir(repo, repo_yml) as path:
                pass
            _set_url_and_fetch(
                repo,
                repo_yml,
                remote_name,
                url,
                filter_remote=filter_remote,
                trycount=trycount + 1,
            )
        else:
            _raise_error(
                f"Even after rebuilding cache dir it was not possible to clone {repo_yml.path}"
            )



import uuid
import os
import subprocess
import click
import shutil
from .gitcommands import GitCommands
from pathlib import Path
from .tools import yieldlist, X, safe_relative_to, _raise_error, rmtree
from .consts import gitcmd as git
from contextlib import contextmanager
from .tools import is_forced
from .tools import temppath
from .tools import filter_files_to_folders
from .tools import split_every


class Repo(GitCommands):
    def __init__(self, path):
        super().__init__(path)
        self.is_submodule = False

    def __repr__(self):
        return f"{self.path}"

    def __str__(self):
        return f"{self.path}"

    def is_path_a_submodule(self, path):
        assert not str(path).startswith("/")
        for subm in self.get_submodules():
            if subm.path == self.path / path:
                return subm
        return False

    def contain_commit(self, commit):
        try:
            self.X(*(git + ["cat-file", "-t", commit]), output=True)
            return True
        except Exception:
            return False

    def contains_branch(self, branch):
        try:
            self.X(
                *(git + ["rev-parse", "--verify", branch]),
                output=True,
            )
            return True
        except Exception:
            return False

    def get_branch(self):
        try:
            res = self.X(
                *(git + ["symbolic-ref", "--short", "HEAD"]),
                output=True,
                env={
                    "PAGER": "",
                    "GIT_PAGER": "",
                },
            )
            return res.splitlines()[0]
        except Exception as ex:
            return None

    def get_commit(self):
        try:
            res = self.X(
                *(git + ["log", "-n", "1", "--format=%H"]),
                output=True,
                env={
                    "PAGER": "",
                    "GIT_PAGER": "",
                },
            )
            return res
        except Exception as ex:
            return False

    def contains(self, commit):
        try:
            self.X(
                *(git + ["branch", "--contains", commit]),
                env={
                    "PAGER": "",
                    "GIT_PAGER": "",
                },
            )
            return True
        except Exception as ex:
            return False

    @property
    def is_bare(self):
        answer = self.out(*(git + ["rev-parse", "--is-bare-repository"]))
        return answer.strip() == "true"

    @property
    def rel_path_to_root_repo(self):
        assert str(self.path).startswith("/")
        return self.path.relative_to(self.root_repo.path)

    @property
    def root_repo(self):
        path = self.path
        for _ in path.parts:
            if (path / ".git").is_dir():
                return Repo(path)
            path = path.parent
        return None

    @property
    def _git_path(self):
        return self.path / ".git"

    def ls_files_states(self, params):
        """
        e.g. ls_files_states("-dmosk" )
        """
        if "-t" not in params:
            params += ["-t"]

        def extract(line):
            if "\t" in line:
                return line.split("\t")[-1]
            return line[2:]

        files = list(
            map(
                lambda line: Path(extract(line)),
                self.out(*(git + ["ls-files"] + params)).splitlines(),
            )
        )
        return files

    def force_remove_submodule(self, path):
        # https://github.com/jeremysears/scripts/blob/master/bin/git-submodule-rewrite
        self.please_no_staged_files()

        # if there are dirty files, then abort to not destroy data
        dirty_files = list(
            filter_files_to_folders(
                self.all_dirty_files_absolute,
                [self.path_absolute / path],
            )
        )
        if dirty_files:
            if not is_forced():
                _raise_error(f"Path is dirty: {path}. Changes would be lost.")
        fullpath = self.path / path
        if fullpath.exists():
            self.X(
                *(
                    git
                    + [
                        "config",
                        "-f",
                        ".gitmodules",
                        "--remove-section",
                        f"submodule.{path}",
                    ]
                ),
                allow_error=True,
            )
            if self.out(
                *(
                    git
                    + [
                        "config",
                        "-f",
                        self.configdir / "config",
                        "--get",
                        f"submodule.{path}.url",
                    ]
                ),
                allow_error=True,
            ):
                self.X(
                    *(
                        git
                        + [
                            "config",
                            "-f",
                            self.configdir / "config",
                            "--remove-section",
                            f"submodule.{path}",
                        ]
                    )
                )

        subrepo = Repo(self.path / path)
        self.X("rm", "-rf", path)
        if fullpath.exists():
            self.X(*(git + ["add", "-A", path]))
        if self.lsfiles(fullpath.relative_to(self.path)):
            self.X(*(git + ["add", "-f", "-A", path]))
        if (self.path_absolute / ".gitmodules") in self.all_dirty_files_absolute:
            self.X(*(git + ["add", "-A", ".gitmodules"]))

        if self.staged_files:
            self.X(*(git + ["commit", "-q", "--no-verify", "-m", f"removed submodule {path}"]))
        self.X("rm", "-rf", f".git/modules/{subrepo.rel_path_to_root_repo}")

    @property
    def next_module_root(self):
        if not self.path.exists():
            return None

        p = self.path
        for i in range(len(p.parts)):
            if (p / ".git").exists():
                return p
            p = p.parent
        return self.path

    def get_submodules_with_new_commits(self):
        submodules = self.out(*(git + ["submodule", "status"])).splitlines()
        for line in submodules:
            splitted = line.strip().split(" ")
            if line.startswith("+") or splitted[1] == "./":
                yield Submodule(
                    self.next_module_root / splitted[1], self.next_module_root
                )

    @yieldlist
    def get_submodules(self):
        submodules = self.out(*(git + ["submodule", "status"])).splitlines()
        # no entry found for x in .gitmodules
        for line in submodules:
            splitted = line.strip().split(" ")
            if line.startswith("-") or splitted[1] == "./":
                # means, that path does not exist right now; path info is misleading ./
                continue
            yield Submodule(self.next_module_root / splitted[1], self.next_module_root)

    def check_ignore(self, path):
        try:
            self.X(*(git + ["check-ignore", "-q", path]), allow_error=False)
        except subprocess.CalledProcessError:
            return False
        else:
            return True

    @contextmanager
    def temporary_unignore(self, path):
        backup_gitignore = None
        gitignore_path = self.path / ".gitignore"
        if gitignore_path.exists(): 
            backup_gitignore = gitignore_path.read_text().splitlines()
        
        try:
            # no index: if file is already committed, then there would be no output
            ignored = self.X(*(git + ["check-ignore", "-v", "--no-index", path]), allow_error=False, output=True)
        except subprocess.CalledProcessError:
            ignored = False
        if ignored and backup_gitignore:
            except_lines = [int(x.split(":", 2)[1]) for x in ignored.splitlines()]
            tempcontent = list(backup_gitignore)
            for except_line in sorted(except_lines, reverse=True):
                del tempcontent[except_line - 1]
            gitignore_path.write_text("\n".join(tempcontent) + "\n")

        yield

        if backup_gitignore:
            gitignore_path.write_text("\n".join(backup_gitignore) + "\n")

    def _fix_to_remove_subdirectories(self, config):
        # https://stackoverflow.com/questions/4185365/no-submodule-mapping-found-in-gitmodule-for-a-path-thats-not-a-submodule
        # commands may block
        # git submodule--helper works and shows something
        # git submodule says: fatal: no submodule mapping found in .gitmodules
        # there is special folder with id 16000 then, then must be removed with git rm
        # then; not tested, because then it suddenly worked
        lines = [
            x
            for x in self.out(*(git + ["ls-files", "--stage"])).splitlines()
            if x.strip().startswith("160000")
        ]
        # 160000 5e8add9536e584f73ea25d4cf51577832d480e90 0       addons_robot
        for line in lines:
            linepath = line.split("\t", 1)[1]
            path = self.path / linepath
            if path.exists():
                from .gimera import REPO_TYPE_SUB

                if not [
                    x
                    for x in config.repos
                    if x.path == path and x.ttype == REPO_TYPE_SUB
                ]:
                    continue
                self.please_no_staged_files()
                # if .gitmodules is dirty then commit that first, otherwise:
                # fatal: please stage your changes to .gitmodules or stash them to proceed
                if (self.path / ".gitmodules") in self.all_dirty_files:
                    self.X(*(git + ["add", ".gitmodules"]))
                    self.X(
                        *(
                            git
                            + [
                                "commit",
                                "-q",
                                "-m",
                                "gimera fix to removed subdirs: .gitmodules",
                            ]
                        )
                    )
                self.X(*(git + ["rm", "-f", linepath]))
                self.X(
                    *(git + ["commit", "-q", "--no-verify", "-m", f"removed invalid subrepo: {linepath}"])
                )

    def get_submodule(self, path):
        for submodule in self.get_submodules():
            if str(submodule.path.relative_to(self.path)) == str(Path(path)):
                return submodule
        raise ValueError(f"Path not found: {path}")

    def fetch(self, remote=None, ref=None):
        if not isinstance(ref, (tuple, list)) and ref:
            ref = [ref]
        elif ref is None:
            ref = []
        if remote:
            if isinstance(remote, str):
                remote = remote
            else:
                remote = remote.name
        else:
            remote = None
        self.X(*(git + ["fetch", remote] + ref + ["--prune"]))

    def fetchall(self):
        self.fetch(remote="origin",ref=["+refs/heads/*:refs/remotes/origin/*","+refs/tags/*:refs/tags/*", "--prune"])

    def get_remote(self, name):
        return [x for x in self.remotes if x.name == name][0]

    def set_remote_url(self, name, url):
        remote = Remote(self, name, url)
        # do not fetch; could point to https form which requires authentication
        self.add_remote(remote, exist_ok=True, no_set_url=True, fetch=False)
        self.X(
            *(git + ["remote", "set-url", name, url]), env={"GIT_TERMINAL_PROMPT": "0"}
        )

    def remove_remote(self, remote):
        self.X(
            *(
                git
                + [
                    "remote",
                    "rm",
                    remote and remote.name or None,
                ]
            ),
            env={"GIT_TERMINAL_PROMPT": "0"},
        )

    def add_remote(self, remote, exist_ok=False, no_set_url=False, fetch=True):
        output = self.out(
            *(git + ["remote"]), env={"GIT_TERMINAL_PROMPT": "0"}
        ).splitlines()
        if [x for x in output if x.strip() == remote.name]:
            if not no_set_url:
                self.set_remote_url(remote.name, remote.url)
        else:
            self.X(
                *(
                    git
                    + [
                        "remote",
                        "add",
                        remote.name,
                        remote.url,
                    ]
                ),
                env={"GIT_TERMINAL_PROMPT": "0"},
            )
        if fetch:
            self.X(*(git + ["fetch", remote.name]), env={"GIT_TERMINAL_PROMPT": "0"})

    def pull(self, remote=None, ref=None, repo_yml=None):
        """
        The git reset hard way was necessary in a live repository; local files
        had to be overridden that stand in the way

        This was too weak:
        self.X("git", "pull", "--no-edit", remote and remote.name or None, ref)

        "params remote": remote object

        """
        if repo_yml:
            assert not remote
            assert not ref
            remote = "origin"
            ref = repo_yml.branch

        if remote and not isinstance(remote, str):
            remote = remote.name

        if not remote and not ref:
            raise Exception("Requires remote and ref or yaml configuration.")

        self.X(*(git + ["pull", "--no-edit", "--no-rebase", remote, ref]))

    def full_clean(self):
        self.X(*(git + ["checkout", "-f"]))
        self.X(*(git + ["clean", "-xdff"]))

    def please_no_staged_files(self):
        staged = self.staged_files
        if not staged:
            return
        _raise_error(
            "For the operation there mustnt be " f"any staged files like {staged}"
        )

    @property
    def remotes(self):
        result = {}
        for line in self.out(*(git + ["remote", "-v"])).splitlines():
            name, url = line.strip().split("\t")
            url = url.split("(")[0].rstrip()
            repo = Remote(self, name, url)
            v = result.setdefault(name, repo)
            if str(v.url) != str(url):
                raise NotImplementedError(
                    (
                        "Different urls for push and fetch for remote {name}\n"
                        f"{url} != {v}"
                    )
                )
            result[name] = repo
        return result.values()

    def clear_empty_subpaths(self, config):
        """
        If subrepo is at ./path1/path2/path3 and it is removed,
        then path1/path2 stays. This leads to dirty files and unneeded files causing
        problems at switch submodules to integrated and back.

        Be careful to integrate gitignore patterns.
        """
        check = self.path / config.path
        dont_go_beyond = self.path / Path(config.path).parts[0]
        while check.exists():
            # removing untracked and ignored files
            # there may also be the case of "excluded" files - never used this,
            # but possible;
            self.X(*(git + ["clean", "-fd", self.path / check]))
            if not safe_relative_to(check, dont_go_beyond):
                break
            if not check.exists():
                check = check.parent
                continue
            if not list(check.iterdir()):
                rmtree(check)
            check = check.parent
            if not safe_relative_to(check, self.path):
                break

    def lsfiles(self, path):
        files = list(map(Path, self.out(*(git + ["ls-files", path])).splitlines()))
        return files

    def commit_dir_if_dirty(self, rel_path, commit_msg, force=False):
        # commit updated directories
        ammend = False
        if any(
            map(
                lambda filepath: safe_relative_to(filepath, self.path / rel_path),
                self.all_dirty_files_absolute,
            )
        ):
            add_cmd = git + ["add"]
            if force:
                add_cmd += ["-f"]
            add_cmd += [rel_path]
            self.X(*add_cmd)
            # if there are no staged files, it can be, that below that, there is a
            # submodule which changed files; then after git add it is not added
            if self.staged_files:
                gitcmd = [
                            "commit",
                            "-q",
                            "--no-verify",
                            "-m",
                            commit_msg,
                        ]
                self.X(
                    *(
                        git
                        + gitcmd
                    )
                )
                ammend = True

                self.run_precommit_if_installed(rel_path, ammend=ammend)

    def run_precommit_if_installed(self, rel_path, ammend=False):
        if os.getenv("GIMERA_NO_PRECOMMIT") == "1":
            return
        if self.is_precommit_used():
            if not shutil.which('pre-commit'):
                print(f"Command 'pre-commit' not found in PATH")
            self.X(
                *tuple(
                    ["pre-commit", "run", "--from-ref", "HEAD~1", "--to-ref", "HEAD"]
                ),
                allow_error=True,
            )

            gitcmd = ["commit", "-q", "--no-verify"]
            if ammend:
                gitcmd += ["--amend", "--no-edit", "--allow-empty"]
            else:
                gitcmd += ["-m", f"pre-commit run for {rel_path}"]

            dirty_files = list(
                filter_files_to_folders(
                    self.all_dirty_files_absolute,
                    [self.path_absolute / rel_path],
                )
            )
            if dirty_files:
                self.X(*(git + ["add", rel_path]))
                self.X(*(git + gitcmd))

    def is_precommit_used(self):
        candidates = [
            self.root_repo.path / ".pre-commit-config.yaml",
            self.root_repo.path / ".pre-commit-config.yml",
        ]
        return any(x.exists() for x in candidates)

    def submodule_add(self, branch, url, rel_path):
        commands = git + [
            "submodule",
            "add",
            "--force",
            "-b",
            str(branch),
            url,
            rel_path,
        ]
        try:
            self.out(*commands)
        except subprocess.CalledProcessError as ex:
            rel_path_repo = safe_relative_to(self.path, self.root_repo.path)
            # self._remove_internal_submodule_clone(self.rel_path_to_root_repo / rel_path)
            self._remove_internal_submodule_clone(rel_path_repo / rel_path.parts[0])
            if (self.path / rel_path).exists():
                rmtree(self.path / rel_path)
            self.out(*commands)

    def _remove_internal_submodule_clone(self, rel_path_to_root):
        # switching integrated/submodule often makes weird conflicts;
        repo = self.root_repo
        root = repo.path / ".git" / "modules"
        # if root.exists():
        #     shutil.rmtree(root)
        #     return
        parts = list(rel_path_to_root.parts)
        next_path = root
        while parts:
            part = parts.pop(0)
            if not parts:
                # kill
                if next_path.exists():
                    rmtree(next_path)
                else:
                    _raise_error(
                        f"Could not delete submodule {part} in {root} - not found"
                    )

            else:
                next_path = next_path / part / "modules"
                if not next_path.exists():
                    next_path = next_path.parent

    @contextmanager
    def stay_at_commit(self, enabled):
        commit = self.hex
        try:
            yield
        finally:
            if enabled:
                self.X(*(git + ["reset", "--soft"]), commit)

    @contextmanager
    def worktree(self, commit):
        with temppath() as tmpfolder:
            repo_folder = tmpfolder / str(uuid.uuid4())
            repo_folder.parent.mkdir(exist_ok=True, parents=True)
            try:
                repo = Repo(repo_folder)
                self.X(*(git + ["worktree", "add", "--force", repo_folder, commit]))
                yield repo
            except Exception as ex:
                click.secho(f"Error occurred at repo {self.path}", fg="red")
                raise
            finally:
                repo.X(*(git + ["worktree", "remove", "--force", repo_folder]))
                rmtree(tmpfolder)

    def move_worktree_content(self, dest_path):
        if dest_path.exists():
            rmtree(dest_path)
        # faster than rsync
        shutil.move(self.path, dest_path)
        gitdir = dest_path / ".git"
        if gitdir.exists():
            self.path.mkdir()
            shutil.move(gitdir, self.path / ".git")


class Remote(object):
    def __init__(self, repo, name, url):
        self.repo = repo
        self.name = name
        self.url = url


class Submodule(Repo):
    def __init__(self, path, parent_path):
        self.path = Path(path)
        self.path_absolute = path.absolute()
        self.parent_path = Path(parent_path)
        self.relpath = self.path.relative_to(self.parent_path)
        self.is_submodule = True

    @property
    def is_git_submodule(self):
        gitmodules = self.parent_path / ".gitmodules"
        if not gitmodules.exists():
            return False
        module = [
            x
            for x in gitmodules.read_text().splitlines()
            if x.startswith("[submodule")
            if f'"{self.relpath}"]' in x
        ]
        return bool(module)

    def __repr__(self):
        return f"{self.path}"

    def __str__(self):
        return f"{self.path}"

    def get_url(self, noerror=True):
        try:
            url = Repo(self.parent_path).out(
                *(
                    git
                    + ["config", "-f", ".gitmodules", f"submodule.{self.relpath}.url"]
                ),
            )
        except subprocess.CalledProcessError:
            if not noerror:
                raise
            url = ""
        return url

    def equals(self, other):
        relpath = str(self.path.relative_to(self.parent_path))
        if isinstance(other, str):
            return relpath == other
        if isinstance(other, Path):
            return self.path.absolute() == other.absolute()
        raise NotImplementedError(other)



from .fixtures import *  # required for all
import itertools
import yaml
from contextlib import contextmanager
from ..repo import Repo
import os
import subprocess
import inspect
import os
from pathlib import Path
from .tools import gimera_apply
from ..tools import rsync
from . import temppath
from .tools import _make_remote_repo
from .tools import clone_and_commit
from .tools import gimera_commit

from ..consts import gitcmd as git
import time

current_dir = Path(
    os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
)




from .fixtures import *  # required for all
import time
import uuid
import yaml
from contextlib import contextmanager
from ..repo import Repo
import os
import subprocess
import click
import inspect
import os
from pathlib import Path
from .tools import gimera_apply
from .tools import _make_remote_repo
from .tools import clone_and_commit

from ..consts import gitcmd as git

current_dir = Path(
    os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
)


def test_submodule_tree_dirty_files(temppath):
    """
    * put same repo integrated and submodule into main repo
    * add file2.txt on remote
    * check after apply that file exists in both
    * make a patch in integrated version
    """
    workspace = temppath / "workspace_tree_dirty_files"
    workspace.mkdir()
    workspace_main = workspace / "main_working"

    repo_main = _make_remote_repo(temppath / "mainrepo")
    repo_sub = _make_remote_repo(temppath / "sub1")
    repo_subsub = _make_remote_repo(temppath / "subsub1")
    repo_2 = _make_remote_repo(temppath / "repo2")

    with clone_and_commit(repo_2, "main") as repopath:
        (repopath / "file1.txt").write_text("This is a new function")
        subprocess.check_call(git + ["add", "file1.txt"], cwd=repopath)
        subprocess.check_call(git + ["commit", "-am", "file1 added"], cwd=repopath)

    with clone_and_commit(repo_subsub, "main") as repopath:
        (repopath / "file1.txt").write_text("This is a new function")
        subprocess.check_call(git + ["add", "file1.txt"], cwd=repopath)
        subprocess.check_call(git + ["commit", "-am", "file1 added"], cwd=repopath)

    with clone_and_commit(repo_sub, "main") as repopath:
        (repopath / "file1.txt").write_text("This is a new function")
        subprocess.check_call(git + ["add", "file1.txt"], cwd=repopath)
        subprocess.check_call(
            git + ["submodule", "add", f"file://{repo_subsub}", "subsub"], cwd=repopath
        )
        subprocess.check_call(git + ["commit", "-am", "file1 added"], cwd=repopath)

    with clone_and_commit(repo_main, "main") as repopath:
        (repopath / "file1.txt").write_text("This is a new function")
        subprocess.check_call(git + ["add", "file1.txt"], cwd=repopath)
        subprocess.check_call(
            git + ["submodule", "add", f"file://{repo_sub}", "sub"], cwd=repopath
        )
        subprocess.check_call(git + ["commit", "-am", "file1 added"], cwd=repopath)

    subprocess.check_output(
        git + ["clone", "file://" + str(repo_main), workspace_main],
        cwd=workspace,
    )
    subprocess.check_call(
        git + ["submodule", "update", "--init", "--recursive"], cwd=workspace_main
    )
    assert (workspace_main / "sub" / "subsub" / "file1.txt").exists()

    from ..gitcommands import GitCommands

    assert not GitCommands(workspace_main / "sub" / "subsub").dirty_existing_files
    assert not GitCommands(workspace_main / "sub" / "subsub").untracked_files
    assert not GitCommands(workspace_main / "sub" / "subsub").all_dirty_files
    (workspace_main / "sub" / "subsub" / "newfile.txt").write_text("new")
    assert not GitCommands(workspace_main / "sub" / "subsub").dirty_existing_files
    assert GitCommands(workspace_main / "sub" / "subsub").untracked_files
    assert GitCommands(workspace_main / "sub" / "subsub").all_dirty_files
    assert GitCommands(workspace_main / "sub").dirty_existing_files
    assert not GitCommands(workspace_main / "sub").untracked_files
    assert GitCommands(workspace_main / "sub").all_dirty_files
    assert GitCommands(workspace_main).dirty_existing_files
    assert not GitCommands(workspace_main).untracked_files
    assert GitCommands(workspace_main).all_dirty_files
    (workspace_main / "sub" / "subsub" / "newfile.txt").unlink()
    assert not GitCommands(workspace_main / "sub" / "subsub").dirty_existing_files
    assert not GitCommands(workspace_main / "sub" / "subsub").untracked_files
    assert not GitCommands(workspace_main / "sub" / "subsub").all_dirty_files
    (workspace_main / "sub" / "newfile.txt").write_text("new")
    assert not GitCommands(workspace_main / "sub" / "subsub").dirty_existing_files
    assert not GitCommands(workspace_main / "sub" / "subsub").untracked_files
    assert not GitCommands(workspace_main / "sub" / "subsub").all_dirty_files
    assert GitCommands(workspace_main).dirty_existing_files
    assert not GitCommands(workspace_main).untracked_files
    assert GitCommands(workspace_main).all_dirty_files
    (workspace_main / "sub" / "newfile.txt").unlink()
    assert not GitCommands(workspace_main / "sub" / "subsub").dirty_existing_files
    assert not GitCommands(workspace_main / "sub" / "subsub").untracked_files
    assert not GitCommands(workspace_main / "sub" / "subsub").all_dirty_files
    (workspace_main / "newfile.txt").write_text("new")
    assert not GitCommands(workspace_main / "sub" / "subsub").dirty_existing_files
    assert not GitCommands(workspace_main / "sub" / "subsub").untracked_files
    assert not GitCommands(workspace_main / "sub" / "subsub").all_dirty_files
    (workspace_main / "newfile.txt").unlink()
    assert not GitCommands(workspace_main / "sub" / "subsub").dirty_existing_files
    assert not GitCommands(workspace_main / "sub" / "subsub").untracked_files
    assert not GitCommands(workspace_main / "sub" / "subsub").all_dirty_files

    # make a submodule and check if marked as dirty
    subprocess.check_call(
        git + ["submodule", "add", f"file://{repo_2}", "repo2"],
        cwd=workspace_main / "sub" / "subsub",
    )
    assert not GitCommands(workspace_main / "sub" / "subsub").dirty_existing_files
    assert GitCommands(workspace_main / "sub" / "subsub").untracked_files
    assert GitCommands(workspace_main / "sub" / "subsub").all_dirty_files

    assert GitCommands(workspace_main / "sub").is_submodule("subsub")


def test_cleanup_dirty_submodule(temppath):
    """
    make dirty submodule then repo.full_clean
    """
    workspace = temppath / "workspace_cleanup_dirty_submodule"
    os.chdir(workspace.parent)

    repo_main = _make_remote_repo(temppath / "mainrepo")
    repo_sub = _make_remote_repo(temppath / "sub1")
    repo_subsub = _make_remote_repo(temppath / "subsub1")

    subprocess.check_output(
        git + ["clone", "file://" + str(repo_main), workspace.name],
        cwd=workspace.parent,
    )
    with clone_and_commit(repo_subsub, "main") as repopath:
        (repopath / "file1.txt").write_text("This is a new function")
        subprocess.check_call(git + ["add", "file1.txt"], cwd=repopath)
        subprocess.check_call(git + ["commit", "-am", "file1 added"], cwd=repopath)

    with clone_and_commit(repo_sub, "main") as repopath:
        (repopath / "file1.txt").write_text("This is a new function")
        subprocess.check_call(git + ["add", "file1.txt"], cwd=repopath)
        subprocess.check_call(
            git + ["submodule", "add", f"file://{repo_subsub}", "subsub"], cwd=repopath
        )
        subprocess.check_call(git + ["commit", "-am", "file1 added"], cwd=repopath)

    with clone_and_commit(repo_main, "main") as repopath:
        (repopath / "file1.txt").write_text("This is a new function")
        subprocess.check_call(git + ["add", "file1.txt"], cwd=repopath)
        subprocess.check_call(
            git + ["submodule", "add", f"file://{repo_sub}", "sub"], cwd=repopath
        )
        subprocess.check_call(git + ["commit", "-am", "file1 added"], cwd=repopath)

    workspace_main = workspace / "main_working"
    subprocess.check_call(git + ["clone", f"file://{repo_main}", workspace_main])
    subprocess.check_call(
        git + ["submodule", "update", "--init", "--recursive"], cwd=workspace_main
    )
    assert (workspace_main / "sub" / "subsub" / "file1.txt").exists()

    # make dirty
    (workspace_main / "sub" / "subsub" / "file5.txt").write_text("data")
    Repo(workspace_main).full_clean()


def test_switch_submodule_to_integrated_and_sub(temppath):
    workspace = temppath / "workspace_switch_submodule"
    workspace.mkdir()
    workspace_main = workspace / "main_working"

    repo_main = _make_remote_repo(temppath / "mainrepo")
    repo_sub = _make_remote_repo(temppath / "sub1")

    repos_sub = {
        "repos": [
            {
                "url": f"file://{repo_sub}",
                "branch": "branch1",
                "path": "sub1",
                "patches": [],
                "type": "submodule",
            },
        ]
    }
    repos_int = {
        "repos": [
            {
                "url": f"file://{repo_sub}",
                "branch": "branch1",
                "path": "sub1",
                "patches": [],
                "type": "integrated",
            },
        ]
    }
    subprocess.check_output(
        git + ["clone", "file://" + str(repo_main), workspace_main],
        cwd=workspace.parent,
    )
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_int))
    (workspace_main / "main.txt").write_text("main repo")
    repo = Repo(workspace_main)
    repo.simple_commit_all()
    repo.X(*(git + ["push"]))

    os.chdir(workspace_main)
    gimera_apply([], None)

    def test_if_change_is_pushed_back():
        file = workspace_main / "sub1" / str(uuid.uuid4())
        file.write_text("content")
        subrepo = repo.get_submodule("sub1")
        subrepo.simple_commit_all()
        subrepo.X(*(git + ["push"]))
        with clone_and_commit(repo_sub, "branch1") as repopath:
            assert (repopath / file.name).exists()

    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_sub))
    os.chdir(workspace_main)
    gimera_apply([], None)
    try:
        repo.get_submodule("sub1")
    except ValueError:
        raise Exception("Should be found")
    test_if_change_is_pushed_back()

    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_int))
    os.chdir(workspace_main)
    repo.X(*(git + ["add", "sub1"]))
    repo.X(*(git + ["commit", "-m", "submodule new revision"]))
    gimera_apply([], None)
    try:
        repo.get_submodule("sub1")
    except ValueError:
        pass
    else:
        raise Exception("Should not be found")

    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_sub))
    os.chdir(workspace_main)
    gimera_apply([], None)
    try:
        repo.get_submodule("sub1")
    except ValueError:
        raise Exception("Should be found")
    test_if_change_is_pushed_back()


def test_switch_submodule_to_integrated_and_sub_with_gitignores(temppath):
    """Took long time to understand it:
    if there are gitignores and a submodule is moved to an integrated module
    then files of that subrepo may become gitignored files of main repo and
    removal of that directory fails.
    For that clear_empty_subpaths is called after switching to git clean
    untracked, ignored and excluded files. Excluded files not really used and tested
    yet.
    """
    workspace = temppath / "workspace_switch_submodule_gitignore"
    workspace.mkdir()
    workspace_main = workspace / "main_working"

    repo_main = _make_remote_repo(temppath / "mainrepo")
    repo_sub = _make_remote_repo(temppath / "sub1")

    repos_sub = {
        "repos": [
            {
                "url": f"file://{repo_sub}",
                "branch": "branch1",
                "path": "sub1",
                "patches": [],
                "type": "submodule",
            },
        ]
    }
    repos_int = {
        "repos": [
            {
                "url": f"file://{repo_sub}",
                "branch": "branch1",
                "path": "sub1",
                "patches": [],
                "type": "integrated",
            },
        ]
    }
    with clone_and_commit(repo_sub, "main") as repopath:
        (repopath / "repo_sub.txt").write_text("This is a new function")
        (repopath / "dont_look_at_me").write_text("i am ugly")
        Repo(repopath).simple_commit_all()

    subprocess.check_output(
        git + ["clone", "file://" + str(repo_main), workspace_main],
        cwd=workspace.parent,
    )
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_int))
    (workspace_main / "main.txt").write_text("main repo")
    (workspace_main / ".gitignore").write_text("dont_look_at_me\n")
    repo = Repo(workspace_main)
    repo.simple_commit_all()
    repo.X(*(git + ["push"]))

    os.chdir(workspace_main)
    gimera_apply([], None)

    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_sub))
    repo.simple_commit_all()
    os.chdir(workspace_main)
    gimera_apply([], None)
    try:
        repo.get_submodule("sub1")
    except ValueError:
        raise Exception("Should be found")
    assert not repo.all_dirty_files

    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_int))
    repo.simple_commit_all()
    os.chdir(workspace_main)
    gimera_apply([], None)
    try:
        repo.get_submodule("sub1")
    except ValueError:
        pass
    else:
        raise Exception("Should not be found")
    assert not repo.all_dirty_files

    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_sub))
    repo.simple_commit_all()
    os.chdir(workspace_main)
    gimera_apply([], None)
    try:
        repo.get_submodule("sub1")
    except ValueError:
        raise Exception("Should be found")
    assert not repo.all_dirty_files


def test_switch_submodule_to_other_url(temppath):
    """
    switch url of sub repo and check if redirected
    """
    workspace = temppath / "workspace_switch_subrepo"
    os.chdir(workspace.parent)

    repo_main = _make_remote_repo(temppath / "mainrepo")
    repo_1 = _make_remote_repo(temppath / "repo1")
    repo_2 = _make_remote_repo(temppath / "repo2")

    subprocess.check_output(
        git + ["clone", "file://" + str(repo_main), workspace.name],
        cwd=workspace.parent,
    )
    with clone_and_commit(repo_1, "main") as repopath:
        (repopath / "repo1.txt").write_text("This is a new function")
        Repo(repopath).simple_commit_all()

    with clone_and_commit(repo_2, "main") as repopath:
        (repopath / "repo2.txt").write_text("This is a new function")
        Repo(repopath).simple_commit_all()

    with clone_and_commit(repo_main, "main") as repopath:
        (repopath / "dummy.txt").write_text("This is a new function")
        Repo(repopath).simple_commit_all()

    workspace_main = workspace / "main_working"
    subprocess.check_call(git + ["clone", f"file://{repo_main}", workspace_main])
    subprocess.check_call(
        git + ["submodule", "update", "--init", "--recursive"], cwd=workspace_main
    )

    repos1 = {
        "repos": [
            {
                "url": f"file://{repo_1}",
                "branch": "main",
                "path": "subby",
                "patches": [],
                "type": "submodule",
            },
        ]
    }
    repos2 = {
        "repos": [
            {
                "url": f"file://{repo_2}",
                "branch": "main",
                "path": "subby",
                "patches": [],
                "type": "submodule",
            },
        ]
    }
    main_repo = Repo(workspace_main)
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos1))
    main_repo.simple_commit_all()
    os.chdir(workspace_main)
    gimera_apply([], None)
    assert (workspace_main / "subby" / "repo1.txt").exists()

    (workspace_main / "gimera.yml").write_text(yaml.dump(repos2))
    main_repo.simple_commit_all()
    os.chdir(workspace_main)
    gimera_apply([], None)
    assert (workspace_main / "subby" / "repo2.txt").exists()


def test_switch_submodule_to_integrated_on_different_branches(temppath):
    """
    switch url of sub repo and check if redirected
    """
    workspace = temppath / "workspace_switch_subrepo"
    os.chdir(workspace.parent)

    repo_main = _make_remote_repo(temppath / "mainrepo")
    repo_1 = _make_remote_repo(temppath / "repo1")

    subprocess.check_output(
        git + ["clone", "file://" + str(repo_main), workspace.name],
        cwd=workspace.parent,
    )
    with clone_and_commit(repo_1, "main") as repopath:
        (repopath / "repo1.txt").write_text("This is a new function")
        Repo(repopath).simple_commit_all()

    with clone_and_commit(repo_main, "main") as repopath:
        (repopath / "dummy.txt").write_text("This is a new function")
        Repo(repopath).simple_commit_all()

    workspace_main = workspace / "main_working"
    subprocess.check_call(git + ["clone", f"file://{repo_main}", workspace_main])
    subprocess.check_call(
        git + ["submodule", "update", "--init", "--recursive"], cwd=workspace_main
    )

    repos = {
        "repos": [
            {
                "url": f"file://{repo_1}",
                "branch": "main",
                "path": "subby",
                "patches": [],
                "type": "submodule",
            },
        ]
    }

    main_repo = Repo(workspace_main)
    main_repo.X(*(git + ["checkout", "-b", "as_submodule"]))
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos))
    main_repo.simple_commit_all()
    os.chdir(workspace_main)
    gimera_apply([], None)
    assert (workspace_main / "subby" / "repo1.txt").exists()
    sha_submodule_step1 = yaml.safe_load((workspace_main / "gimera.yml").read_text())[
        "repos"
    ][0]["sha"]
    assert sha_submodule_step1

    main_repo.X(*(git + ["checkout", "-b", "as_integrated"]))
    repos["repos"][0]["type"] = "integrated"
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos))
    main_repo.simple_commit_all()
    os.chdir(workspace_main)
    gimera_apply([], None)
    assert (workspace_main / "subby" / "repo1.txt").exists()
    assert not (workspace_main / "subby" / "repo2.txt").exists()
    assert not (workspace_main / ".gitmodules").read_text()

    main_repo.X(*(git + ["checkout", "as_submodule"]))
    assert (workspace_main / ".gitmodules").read_text()

    with clone_and_commit(repo_1, "main") as repopath:
        (repopath / "repo2.txt").write_text("This is a new function")
        Repo(repopath).simple_commit_all()

    os.chdir(workspace_main)
    assert yaml.safe_load((workspace_main / "gimera.yml").read_text())["repos"][0][
        "sha"
    ]
    gimera_apply([], False)
    assert (workspace_main / "subby" / "repo1.txt").exists()
    assert not (workspace_main / "subby" / "repo2.txt").exists()

    os.chdir(workspace_main)
    os.environ["GIMERA_NON_THREADED"] = "1"
    gimera_apply([], True)
    assert (workspace_main / "subby" / "repo1.txt").exists()
    assert (workspace_main / "subby" / "repo2.txt").exists()

    main_repo.X(*(git + ["checkout", "-f", "as_integrated"]))
    main_repo.X("git", "clean", "-xdff")
    gimera_apply([], False)
    assert (workspace_main / "subby" / "repo1.txt").exists()
    assert not (workspace_main / "subby" / "repo2.txt").exists()


def test_merges(temppath):
    workspace = temppath / "workspace_switch_subrepo"
    os.chdir(workspace.parent)

    repo_main = _make_remote_repo(temppath / "mainrepo")
    repo_1 = _make_remote_repo(temppath / "repo1")
    repo_1variant = temppath / "repo1variant"

    subprocess.check_output(
        git + ["clone", "file://" + str(repo_main), workspace.name],
        cwd=workspace.parent,
    )
    # make variant repo and change on a branch there
    with clone_and_commit(repo_1, "main") as repopath:
        (repopath / "repo1.txt").write_text("This is a new function")
        subprocess.check_call(["sync"])
        repo = Repo(repopath)
        repo.simple_commit_all()
        repo.X(*(git + ["push"]))

    subprocess.check_call(git + ["clone", "--mirror", "--bare", repo_1, repo_1variant])
    subprocess.check_call(["sync"])

    with clone_and_commit(repo_1variant, "main") as repopath_variant:
        assert Path(repopath_variant / "repo1.txt").exists()
        variant = Repo(repopath_variant)
        variant.X(*(git + ["checkout", "-b", "variant"]))
        (repopath_variant / "variant.txt").write_text("This is a new function")
        variant.simple_commit_all()
        variant.X(*(git + ["push", "--set-upstream", "origin", "variant"]))
        variant.X(*(git + ["checkout", "main"]))

        variant.X(*(git + ["checkout", "-b", "variant2"]))
        (repopath_variant / "variant2.txt").write_text("This is a new function")
        variant.simple_commit_all()
        variant.X(*(git + ["push", "--set-upstream", "origin", "variant2"]))
        variant.X(*(git + ["checkout", "main"]))

    with clone_and_commit(repo_main, "main") as repopath:
        (repopath / "dummy.txt").write_text("This is a new function")
        Repo(repopath).simple_commit_all()

    workspace_main = workspace / "main_working"
    subprocess.check_call(git + ["clone", f"file://{repo_main}", workspace_main])
    subprocess.check_call(
        git + ["submodule", "update", "--init", "--recursive"], cwd=workspace_main
    )

    repos = {
        "repos": [
            {
                "url": f"file://{repo_1}",
                "branch": "main",
                "path": "subby",
                "remotes": {
                    "repo_variant": str(repo_1variant),
                },
                "merges": ["repo_variant variant"],
                "patches": [],
                "type": "integrated",
            },
        ]
    }

    main_repo = Repo(workspace_main)
    main_repo.X(*(git + ["checkout", "-b", "as_submodule"]))
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos))
    main_repo.simple_commit_all()
    subprocess.check_call(["sync"])
    os.chdir(workspace_main)
    gimera_apply([], None)
    assert (workspace_main / "subby" / "variant.txt").exists()
    assert (workspace_main / "subby" / "repo1.txt").exists()

    # reapply should also work
    gimera_apply([], None)
    assert (workspace_main / "subby" / "variant.txt").exists()
    assert (workspace_main / "subby" / "repo1.txt").exists()

    # change the merge - no changes from other merge should exist
    repos = {
        "repos": [
            {
                "url": f"file://{repo_1}",
                "branch": "main",
                "path": "subby",
                "remotes": {
                    "repo_variant": str(repo_1variant),
                },
                "merges": ["repo_variant variant2"],
                "patches": [],
                "type": "integrated",
            },
        ]
    }
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos))
    main_repo.simple_commit_all()
    gimera_apply([], None)
    assert not (workspace_main / "subby" / "variant.txt").exists()
    assert (workspace_main / "subby" / "variant2.txt").exists()
    assert (workspace_main / "subby" / "repo1.txt").exists()


def test_clean_a_submodule_in_submodule(temppath):
    workspace = temppath / "workspace_switch_subrepo"
    os.chdir(workspace.parent)

    repo_main = _make_remote_repo(temppath / "mainrepo")
    repo_1 = _make_remote_repo(temppath / "repo1")
    repo_2 = _make_remote_repo(temppath / "repo1variant")

    subprocess.check_output(
        git + ["clone", "file://" + str(repo_main), workspace.name],
        cwd=workspace.parent,
    )
    # make variant repo and change on a branch there
    with clone_and_commit(repo_2, "main") as repopath:
        (repopath / "repo2.txt").write_text("This is a new function")
        Repo(repopath).simple_commit_all()

    with clone_and_commit(repo_1, "main") as repopath:
        (repopath / "repo1.txt").write_text("This is a new function")
        Repo(repopath).X(
            *(git + ["submodule", "add", f"file://{repo_2}", "folder_of_repo2/repo2"])
        )
        Repo(repopath).simple_commit_all()

    with clone_and_commit(repo_main, "main") as repopath:
        (repopath / "dummy.txt").write_text("This is a new function")
        Repo(repopath).X(
            *(git + ["submodule", "add", f"file://{repo_1}", "folder_of_repo1"])
        )
        Repo(repopath).simple_commit_all()

    workspace_main = workspace / "main_working"
    subprocess.check_call(git + ["clone", f"file://{repo_main}", workspace_main])
    subprocess.check_call(
        git + ["submodule", "update", "--init", "--recursive"], cwd=workspace_main
    )
    main_repo = Repo(workspace_main)
    assert (workspace_main / "folder_of_repo1" / "repo1.txt").exists()
    assert (
        workspace_main / "folder_of_repo1" / "folder_of_repo2" / "repo2" / "repo2.txt"
    ).exists()

    submodule_repo1 = main_repo.get_submodule("folder_of_repo1")
    submodule_repo1.force_remove_submodule("folder_of_repo2/repo2")


def test_switch_submodule_to_integrated_and_sub_with_gitignoring_main_repo(temppath):
    workspace = temppath / "workspace_switch_submodule_gitignore"
    workspace.mkdir()
    workspace_main = workspace / "main_working"

    repo_main = _make_remote_repo(temppath / "mainrepo")
    repo_sub = _make_remote_repo(temppath / "sub1")

    repos_sub = {
        "repos": [
            {
                "url": f"file://{repo_sub}",
                "branch": "branch1",
                "path": "sub1",
                "patches": [],
                "type": "submodule",
            },
        ]
    }
    repos_int = {
        "repos": [
            {
                "url": f"file://{repo_sub}",
                "branch": "branch1",
                "path": "sub1",
                "patches": [],
                "type": "integrated",
            },
        ]
    }
    with clone_and_commit(repo_sub, "main") as repopath:
        (repopath / "repo_sub.txt").write_text("This is a new function")
        (repopath / "dont_look_at_me").write_text("i am ugly")
        Repo(repopath).simple_commit_all()

    subprocess.check_output(
        git + ["clone", "file://" + str(repo_main), workspace_main],
        cwd=workspace.parent,
    )
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_int))
    (workspace_main / "main.txt").write_text("main repo")
    (workspace_main / ".gitignore").write_text("dont_look_at_me\n")
    repo = Repo(workspace_main)
    repo.simple_commit_all()
    repo.X(*(git + ["push"]))

    os.chdir(workspace_main)

    gitignore = workspace_main / ".gitignore"
    gitignore.write_text("sub1")
    gimera_apply([], None)

    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_sub))
    repo.simple_commit_all()
    os.chdir(workspace_main)
    os.environ["BREAKPOINT"] = "1"
    gimera_apply([], None)
    try:
        repo.get_submodule("sub1")
    except ValueError:
        raise Exception("Should be found")
    assert not repo.all_dirty_files

    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_int))
    repo.simple_commit_all()
    os.chdir(workspace_main)
    gimera_apply([], None)
    try:
        repo.get_submodule("sub1")
    except ValueError:
        pass
    else:
        raise Exception("Should not be found")
    assert not repo.all_dirty_files

    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_sub))
    repo.simple_commit_all()
    os.chdir(workspace_main)
    gimera_apply([], None)
    try:
        repo.get_submodule("sub1")
    except ValueError:
        raise Exception("Should be found")
    assert not repo.all_dirty_files


def test_git_submodule_point_to_branch_if_last_commit_matches_tip_point_of_branch(
    temppath,
):
    workspace = temppath / "workspace_switch_submodule_gitignore"
    workspace.mkdir()
    workspace_main = workspace / "main_working"

    repo_main = _make_remote_repo(temppath / "mainrepo")
    repo_sub = _make_remote_repo(temppath / "sub1")

    repos_sub = {
        "repos": [
            {
                "url": f"file://{repo_sub}",
                "branch": "main",
                "path": "sub1",
                "patches": [],
                "type": "submodule",
            },
        ]
    }
    with clone_and_commit(repo_sub, "main") as repopath:
        (repopath / "repo_sub.txt").write_text("This is a new function")
        Repo(repopath).simple_commit_all()

    subprocess.check_output(
        git + ["clone", "file://" + str(repo_main), workspace_main],
        cwd=workspace.parent,
    )
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_sub))
    (workspace_main / "main.txt").write_text("main repo")
    (workspace_main / ".gitignore").write_text("dont_look_at_me\n")
    repo = Repo(workspace_main)
    repo.simple_commit_all()
    repo.X(*(git + ["push"]))
    os.chdir(workspace_main)
    gimera_apply([], None)

    # now update the submodule and remember commit sha
    with clone_and_commit(repo_sub, "main") as repopath:
        (repopath / "repo_sub.txt").write_text("This is a new function2")
        reposub = Repo(repopath)
        reposub.simple_commit_all()
        commit = reposub.get_commit()
        branch = reposub.get_branch()
        assert branch == "main"

    repos_sub["repos"][0]["sha"] = commit
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_sub))
    os.chdir(workspace_main)
    gimera_apply([], {})

    submodule = repo.get_submodule("sub1")
    assert submodule.get_commit() == commit
    assert submodule.get_branch() == branch, "Should now be on the branch, not a sha"


def test_2_submodules(temppath):
    """
    The delete submodule was too hard and deleted submodules
    """
    workspace = temppath / "workspace_2submodules"
    workspace.mkdir(exist_ok=True)
    os.chdir(workspace.parent)

    repo_main = _make_remote_repo(workspace / "mainrepo")
    repo_1 = _make_remote_repo(workspace / "repo1")
    subrepo_1 = _make_remote_repo(workspace / "subrepo1")
    repo_2 = _make_remote_repo(workspace / "repo2")
    subrepo_2 = _make_remote_repo(workspace / "subrepo2")

    with clone_and_commit(subrepo_1, "main") as repopath:
        (repopath / "subrepo1.txt").write_text("This is a new function")
        Repo(repopath).simple_commit_all()
    with clone_and_commit(subrepo_2, "main") as repopath:
        (repopath / "subrepo2.txt").write_text("This is a new function")
        Repo(repopath).simple_commit_all()

    with clone_and_commit(repo_1, "main") as repopath:
        (repopath / "repo1.txt").write_text("This is a new function")
        (repopath / "gimera.yml").write_text(
            yaml.dump(
                {
                    "repos": [
                        {
                            "url": f"file://{subrepo_1}",
                            "branch": "main",
                            "path": "subrepo1",
                            "type": "submodule",
                        },
                    ]
                }
            )
        )
        Repo(repopath).simple_commit_all()

    with clone_and_commit(repo_2, "main") as repopath:
        (repopath / "repo2.txt").write_text("This is a new function")
        (repopath / "gimera.yml").write_text(
            yaml.dump(
                {
                    "repos": [
                        {
                            "url": f"file://{subrepo_2}",
                            "branch": "main",
                            "path": "subrepo2",
                            "type": "submodule",
                        },
                    ]
                }
            )
        )
        Repo(repopath).simple_commit_all()

    repos = {
        "repos": [
            {
                "url": f"file://{repo_1}",
                "branch": "main",
                "path": "repo1",
                "type": "submodule",
            },
            {
                "url": f"file://{repo_2}",
                "branch": "main",
                "path": "repo2",
                "type": "submodule",
            },
        ]
    }

    workspace_main = workspace / "main_working"
    main_repo = Repo(workspace_main)
    subprocess.check_call(git + ["clone", f"file://{repo_main}", workspace_main])
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos))

    os.chdir(workspace_main)
    gimera_apply([], update=None, recursive=True, strict=True)
    assert (workspace_main / "repo1" / "repo1.txt").exists()
    assert (workspace_main / "repo1" / "subrepo1" / "subrepo1.txt").exists()
    assert (workspace_main / "repo2" / "repo2.txt").exists()
    assert (workspace_main / "repo2" / "subrepo2" / "subrepo2.txt").exists()

    submodule_repo1 = main_repo.get_submodule("repo1")
    submodule_repo2 = main_repo.get_submodule("repo2")
    assert submodule_repo1
    assert submodule_repo2



from .fixtures import * # required for all
import uuid
import yaml
from contextlib import contextmanager
from ..repo import Repo
import os
import subprocess
import click
import inspect
import os
from pathlib import Path
from .tools import gimera_apply
from . import temppath
from .tools import _make_remote_repo
from .tools import clone_and_commit
from .tools import gimera_commit

from ..consts import gitcmd as git


def test_commit(temppath):
    """
    * if odoo path is ignored, it is cool to make a patch for it, too
    """
    workspace = temppath / "workspace"

    remote_main_repo = _make_remote_repo(temppath / "mainrepo")
    remote_sub_repo = _make_remote_repo(temppath / "sub1")

    subprocess.check_output(
        git + ["clone", "file://" + str(remote_main_repo), workspace.name],
        cwd=workspace.parent,
    )
    os.environ["GIMERA_NON_INTERACTIVE"] = "1"

    # region gimera config
    repos = {
        "repos": [
            {
                "url": f"file://{remote_sub_repo}",
                "branch": "branch1",
                "path": "sub1",
                "type": "integrated",
            },
        ]
    }
    # endregion

    (workspace / "gimera.yml").write_text(yaml.dump(repos))
    (workspace / "main.txt").write_text("main repo")
    subprocess.check_call(git + ["add", "main.txt"], cwd=workspace)
    subprocess.check_call(git + ["add", "gimera.yml"], cwd=workspace)
    subprocess.check_call(git + ["commit", "-am", "on main"], cwd=workspace)
    subprocess.check_call(git + ["push"], cwd=workspace)
    os.chdir(workspace)
    gimera_apply([], None)
    Repo(workspace).simple_commit_all()
    assert not Repo(workspace).staged_files

    click.secho(
        "Now we have a repo with integrated and gitignored sub"
        "\nWe change something and check if a patch is made."
    )
    os.chdir(workspace)
    os.environ["GIMERA_NON_INTERACTIVE"] = "1"
    os.environ["GIMERA_EXCEPTION_THAN_SYSEXIT"] = "1"

    os.chdir(workspace)
    gimera_apply([], update=False)

    (workspace / "sub1" / "file2.txt").write_text("a new file!")

    gimera_commit("sub1", "branch1", "i committed", False)
    os.environ["GIMERA_FORCE"] = "1"
    os.unlink(workspace / "sub1" / "file2.txt")
    subprocess.check_output(git + ["checkout", "sub1/file1.txt"])
    os.environ["GIMERA_NON_THREADED"] = "1"
    gimera_apply([], update=True)
    assert (
        workspace / "sub1" / "file2.txt"
    ).exists(), "sub1/file2.txt should now exist"

    # failed patch handling
    file1 = workspace / "sub1" / "file1.txt"
    file1.write_text("main\na change!")
    gimera_commit("sub1", "branch1", "i committed", False)
    subprocess.check_output(git + ["checkout", "sub1/file1.txt"])
    gimera_apply([], update=True)
    assert "a change!" in (workspace / "sub1" / "file1.txt").read_text()



from .fixtures import * # required for all
import itertools
import uuid
import yaml
from contextlib import contextmanager
from ..repo import Repo
import os
import subprocess
from pathlib import Path
import shutil
import click
import inspect
import os
from .tools import gimera_apply
from . import temppath
from .tools import _make_remote_repo
from .tools import clone_and_commit
from .tools import gimera_commit

from ..consts import gitcmd as git


def test_common_patchfiles_in_subgimera(temppath):
    """
    * put same repo integrated and submodule into main repo
    * add file2.txt on remote
    * check after apply that file exists in both
    * make a patch in integrated version
    """
    workspace = temppath / "workspace"

    remote_main_repo = _make_remote_repo(temppath / "mainrepo")

    subprocess.check_output(
        git + ["clone", "file://" + str(remote_main_repo), workspace.name],
        cwd=workspace.parent,
    )
    os.environ["GIMERA_NON_INTERACTIVE"] = "1"

    # Make a repo with a patch file and gimera instruction file to
    # include the local patch files depending on the variable $VERSION
    with clone_and_commit(remote_main_repo, "branch1", commit=False) as repopath:
        file1 = repopath / "file_is_patch.txt"
        file1.write_text("patchfile")

        repo = Repo(repopath)
        repo.simple_commit_all()
        patch_content = subprocess.check_output(
            ["git", "format-patch", "HEAD~1", "--stdout", "--relative"],
            encoding="utf8",
            cwd=repopath,
        )

    # region gimera config
    repos = {
        "common": {
            "vars": {"branch": "branch1"},
            "patches": [],
        },
        "repos": [
            {
                "url": f"file://{remote_main_repo}",
                "branch": "${branch}",
                "path": "integrated/sub1",
                "type": "integrated",
                "patches": [
                    {"path": "mypatches"}
                ]
            },
        ],
    }
    # endregion

    (workspace / "gimera.yml").write_text(yaml.dump(repos))
    (workspace / "main.txt").write_text("main repo")
    (workspace / "mypatches").mkdir(parents=True, exist_ok=True)
    (workspace / "mypatches" / "patch1.patch").write_text(patch_content)
    subprocess.check_call(git + ["add", "mypatches"], cwd=workspace)
    subprocess.check_call(git + ["add", "main.txt"], cwd=workspace)
    subprocess.check_call(git + ["add", "gimera.yml"], cwd=workspace)
    subprocess.check_call(git + ["commit", "-am", "on main"], cwd=workspace)
    subprocess.check_call(git + ["push"], cwd=workspace)
    os.chdir(workspace)
    gimera_apply([], None)
    subprocess.check_call(git + ["add", "gimera.yml"], cwd=workspace)
    assert not Repo(workspace).staged_files

    click.secho(
        "Now we have a repo with two subrepos; now we update the subrepos and pull"
    )

    os.chdir(workspace)
    os.environ["GIMERA_NON_INTERACTIVE"] = "1"
    os.environ["GIMERA_EXCEPTION_THAN_SYSEXIT"] = "1"
    gimera_apply([], update=True, recursive=True, strict=True)
    testfile = workspace / "integrated" / "sub1" / "file_is_patch.txt"
    assert testfile.exists()

    # ignore patchfile now
    patchfile = list(
        (workspace / "mypatches").rglob("*.patch")
    )[0]
    repos["repos"][0]["ignored_patchfiles"] = [patchfile.name]
    (workspace / "gimera.yml").write_text(yaml.dump(repos))
    gimera_apply([], None)
    assert not testfile.exists()




import click

class RuntimeState:
    mode = None
    flags = {}

runtime_state = {
    'temppaths': {
    }
}

from . import consts
from . import gimera
from . import repo
from . import gitcommands
from . import tools
from . import snapshot

